/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 *  Copyright (C) 1996  Ralf Baechle
 */
#include <asm/bootinfo.h>
#include <asm/cachectl.h>
#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/system.h>

extern asmlinkage void mips3_cacheflush(void *addr, int nbytes, unsigned int flags);

unsigned long page_colour_mask;

void (*mips_cache_init)(void);

static void
mips3_cache_init(void)
{
	extern asmlinkage void handle_vcei(void);
	extern asmlinkage void handle_vced(void);
	int i;

	switch (mips_cputype) {
	case CPU_R4000MC: case CPU_R4400MC:
	case CPU_R4000SC: case CPU_R4400SC:
		/*
		 * Handlers not implemented yet.  Shouldn't ever be used -
		 * otherwise it's a bug in the Linux/MIPS kernel, anyway.
		 */
		set_except_vector(14, handle_vcei);
		set_except_vector(31, handle_vced);
		break;
	default:
	}

	/*
	 * Compute mask for page_colour().  This is based on the
	 * size of the data cache.
	 */
	i = read_32bit_cp0_register(CP0_CONFIG);
	i = (i >> 26) & 7;
	page_colour_mask = 1 << (12 + i);
	cacheflush = mips3_cacheflush;
}

void (*switch_to_user_mode)(struct pt_regs *regs);

static void
mips3_switch_to_user_mode(struct pt_regs *regs)
{
	regs->cp0_status = (regs->cp0_status & ~(ST0_CU0|ST0_KSU)) | KSU_USER;
}

void
mips3_cpu_init(void)
{
	mips_cache_init = mips3_cache_init;
	switch_to_user_mode = mips3_switch_to_user_mode;
}
